@@ -1,16 +1,24 @@
 //! Extract links and fragments from markdown documents
 use std::collections::{HashMap, HashSet};
 
-use pulldown_cmark::{CowStr, Event, LinkType, Options, Parser, Tag, TagEnd, TextMergeStream};
+use pulldown_cmark::{CowStr, Event, LinkType, Options, Parser, Tag, TagEnd, TextMergeWithOffset};
 
-use crate::{extract::plaintext::extract_raw_uri_from_plaintext, types::uri::raw::RawUri};
+use crate::{
+    extract::{html::html5gum::extract_html_with_span, plaintext::extract_raw_uri_from_plaintext},
+    types::uri::raw::{
+        OffsetSpanProvider, RawUri, RawUriSpan, SourceSpanProvider, SpanProvider as _,
+    },
+};
 
-use super::html::html5gum::{extract_html, extract_html_fragments};
+use super::html::html5gum::extract_html_fragments;
 
 /// Returns the default markdown extensions used by lychee.
 /// Sadly, `|` is not const for `Options` so we can't use a const global.
 fn md_extensions() -> Options {
-    Options::ENABLE_HEADING_ATTRIBUTES | Options::ENABLE_MATH | Options::ENABLE_WIKILINKS
+    Options::ENABLE_HEADING_ATTRIBUTES
+        | Options::ENABLE_MATH
+        | Options::ENABLE_WIKILINKS
+        | Options::ENABLE_FOOTNOTES
 }
 
 /// Extract unparsed URL strings from a Markdown string.
@@ -23,10 +31,13 @@ pub(crate) fn extract_markdown(
     // which is why we keep track of entries and exits while traversing the input.
     let mut inside_code_block = false;
     let mut inside_link_block = false;
+    let mut inside_wikilink_block = false;
 
-    let parser = TextMergeStream::new(Parser::new_ext(input, md_extensions()));
+    let span_provider = SourceSpanProvider::from_input(input);
+    let parser =
+        TextMergeWithOffset::new(Parser::new_ext(input, md_extensions()).into_offset_iter());
     parser
-        .filter_map(|event| match event {
+        .filter_map(|(event, span)| match event {
             // A link.
             Event::Start(Tag::Link {
                 link_type,
@@ -40,14 +51,8 @@ pub(crate) fn extract_markdown(
                     // Inline link like `[foo](bar)`
                     // This is the most common link type
                     LinkType::Inline => {
-                        Some(vec![RawUri {
-                            text: dest_url.to_string(),
-                            // Emulate `<a href="...">` tag here to be compatible with
-                            // HTML links. We might consider using the actual Markdown
-                            // `LinkType` for better granularity in the future
-                            element: Some("a".to_string()),
-                            attribute: Some("href".to_string()),
-                        }])
+                        inside_link_block = true;
+                        Some(raw_uri(&dest_url, span_provider.span(span.start)))
                     }
                     // Reference without destination in the document, but resolved by the `broken_link_callback`
                     LinkType::Reference |
@@ -60,44 +65,38 @@ pub(crate) fn extract_markdown(
                     // Shortcut link like `[foo]`
                     LinkType::Shortcut |
                     // Shortcut without destination in the document, but resolved by the `broken_link_callback`
-                    LinkType::ShortcutUnknown |
+                    LinkType::ShortcutUnknown => {
+                        inside_link_block = true;
+                        // For reference links, create RawUri directly to handle relative file paths
+                        // that linkify doesn't recognize as URLs
+                        Some(raw_uri(&dest_url, span_provider.span(span.start)))
+                    },
                     // Autolink like `<http://foo.bar/baz>`
                     LinkType::Autolink |
                     // Email address in autolink like `<john@example.org>`
-                    LinkType::Email =>
-                     Some(extract_raw_uri_from_plaintext(&dest_url)),
+                    LinkType::Email => {
+                        let span_provider = get_email_span_provider(&span_provider, &span, link_type);
+                        Some(extract_raw_uri_from_plaintext(&dest_url, &span_provider))
+                    }
                     // Wiki URL (`[[http://example.com]]`)
                     LinkType::WikiLink { has_pothole: _ } => {
                         // Exclude WikiLinks if not explicitly enabled
                         if !include_wikilinks {
                             return None;
                         }
-                        inside_link_block = true;
-                        //Ignore gitlab toc notation: https://docs.gitlab.com/user/markdown/#table-of-contents
+                        inside_wikilink_block = true;
+                        // Ignore gitlab toc notation: https://docs.gitlab.com/user/markdown/#table-of-contents
                         if ["_TOC_".to_string(), "TOC".to_string()].contains(&dest_url.to_string()) {
                             return None;
                         }
-                        Some(vec![RawUri {
-                            text: dest_url.to_string(),
-                            element: Some("a".to_string()),
-                            attribute: Some("href".to_string()),
-                        }])
+
+                        // wiki links start with `[[`, so offset the span by `2`
+                        Some(raw_uri(&dest_url, span_provider.span(span.start + 2)))
                     }
                 }
             }
 
-            // An image.
-            // The first field is the link type, the second the destination URL and the third is a title.
-            Event::Start(Tag::Image { dest_url, .. }) => {
-                Some(vec![RawUri {
-                    text: dest_url.to_string(),
-                    // Emulate `<img src="...">` tag here to be compatible with
-                    // HTML links. We might consider using the actual Markdown
-                    // `LinkType` for better granularity in the future
-                    element: Some("img".to_string()),
-                    attribute: Some("src".to_string()),
-                }])
-            }
+            Event::Start(Tag::Image { dest_url, .. }) => Some(extract_image(&dest_url, span_provider.span(span.start))),
 
             // A code block (inline or fenced).
             Event::Start(Tag::CodeBlock(_)) => {
@@ -111,10 +110,15 @@ pub(crate) fn extract_markdown(
 
             // A text node.
             Event::Text(txt) => {
-                if (inside_code_block && !include_verbatim) || inside_link_block {
+                if inside_wikilink_block
+                    || (inside_link_block && !include_verbatim)
+                    || (inside_code_block && !include_verbatim) {
                     None
                 } else {
-                    Some(extract_raw_uri_from_plaintext(&txt))
+                    Some(extract_raw_uri_from_plaintext(
+                        &txt,
+                        &OffsetSpanProvider { offset: span.start, inner: &span_provider }
+                    ))
                 }
             }
 
@@ -122,24 +126,36 @@ pub(crate) fn extract_markdown(
             Event::Html(html) | Event::InlineHtml(html) => {
                 // This won't exclude verbatim links right now, because HTML gets passed in chunks
                 // by pulldown_cmark. So excluding `<pre>` and `<code>` is not handled right now.
-                Some(extract_html(&html, include_verbatim))
+                Some(extract_html_with_span(
+                    &html,
+                    include_verbatim,
+                    OffsetSpanProvider { offset: span.start, inner: &span_provider }
+                ))
             }
 
             // An inline code node.
             Event::Code(code) => {
                 if include_verbatim {
-                    Some(extract_raw_uri_from_plaintext(&code))
+                    // inline code starts with '`', so offset the span by `1`.
+                    Some(extract_raw_uri_from_plaintext(
+                        &code,
+                        &OffsetSpanProvider { offset: span.start + 1, inner: &span_provider }
+                    ))
                 } else {
                     None
                 }
             }
 
-            // A detected link block.
             Event::End(TagEnd::Link) => {
                 inside_link_block = false;
+                inside_wikilink_block = false;
                 None
             }
 
+            // Skip footnote references and definitions explicitly - they're not links to check
+            #[allow(clippy::match_same_arms)]
+            Event::FootnoteReference(_) | Event::Start(Tag::FootnoteDefinition(_)) | Event::End(TagEnd::FootnoteDefinition) => None,
+
             // Silently skip over other events
             _ => None,
         })
@@ -147,6 +163,56 @@ pub(crate) fn extract_markdown(
         .collect()
 }
 
+fn get_email_span_provider<'a>(
+    span_provider: &'a SourceSpanProvider<'_>,
+    span: &std::ops::Range<usize>,
+    link_type: LinkType,
+) -> OffsetSpanProvider<'a> {
+    let offset = match link_type {
+        // We don't know how the link starts, so don't offset the span.
+        LinkType::Reference | LinkType::CollapsedUnknown | LinkType::ShortcutUnknown => 0,
+        // These start all with `[` or `<`, so offset the span by `1`.
+        LinkType::ReferenceUnknown
+        | LinkType::Collapsed
+        | LinkType::Shortcut
+        | LinkType::Autolink
+        | LinkType::Email => 1,
+        _ => {
+            debug_assert!(false, "Unexpected email link type: {link_type:?}");
+            0
+        }
+    };
+
+    OffsetSpanProvider {
+        offset: span.start + offset,
+        inner: span_provider,
+    }
+}
+
+/// Emulate `<img src="...">` tag to be compatible with HTML links.
+/// We might consider using the actual Markdown `LinkType` for better granularity in the future.
+fn extract_image(dest_url: &CowStr<'_>, span: RawUriSpan) -> Vec<RawUri> {
+    vec![RawUri {
+        text: dest_url.to_string(),
+        element: Some("img".to_string()),
+        attribute: Some("src".to_string()),
+        span,
+    }]
+}
+
+/// Emulate `<a href="...">` tag to be compatible with HTML links.
+/// We might consider using the actual Markdown `LinkType` for better granularity in the future.
+fn raw_uri(dest_url: &CowStr<'_>, span: RawUriSpan) -> Vec<RawUri> {
+    vec![RawUri {
+        text: dest_url.to_string(),
+        element: Some("a".to_string()),
+        attribute: Some("href".to_string()),
+        // Sadly, we don't know how long the `foo` part in `[foo](bar)` is,
+        // so the span points to the `[` and not to the `b`.
+        span,
+    }]
+}
+
 /// Extract fragments/anchors from a Markdown string.
 ///
 /// Fragments are generated from headings using the same unique kebab case method as GitHub.
@@ -236,6 +302,8 @@ impl HeadingIdGenerator {
 
 #[cfg(test)]
 mod tests {
+    use crate::types::uri::raw::span;
+
     use super::*;
 
     const MD_INPUT: &str = r#"
@@ -280,11 +348,13 @@ or inline like `https://bar.org` for instance.
                 text: "https://foo.com".to_string(),
                 element: Some("a".to_string()),
                 attribute: Some("href".to_string()),
+                span: span(4, 19),
             },
             RawUri {
                 text: "http://example.com".to_string(),
                 element: Some("a".to_string()),
                 attribute: Some("href".to_string()),
+                span: span(18, 1),
             },
         ];
 
@@ -299,21 +369,25 @@ or inline like `https://bar.org` for instance.
                 text: "https://foo.com".to_string(),
                 element: Some("a".to_string()),
                 attribute: Some("href".to_string()),
+                span: span(4, 19),
             },
             RawUri {
                 text: "https://bar.com/123".to_string(),
                 element: None,
                 attribute: None,
+                span: span(11, 1),
             },
             RawUri {
                 text: "https://bar.org".to_string(),
                 element: None,
                 attribute: None,
+                span: span(14, 17),
             },
             RawUri {
                 text: "http://example.com".to_string(),
                 element: Some("a".to_string()),
                 attribute: Some("href".to_string()),
+                span: span(18, 1),
             },
         ];
 
@@ -385,6 +459,7 @@ $$
             text: "https://example.com/_/foo".to_string(),
             element: None,
             attribute: None,
+            span: span(1, 1),
         }];
         let uris = extract_markdown(markdown, true, false);
         assert_eq!(uris, expected);
@@ -397,6 +472,7 @@ $$
             text: "https://example.com/_".to_string(),
             element: None,
             attribute: None,
+            span: span(1, 1),
         }];
         let uris = extract_markdown(markdown, true, false);
         assert_eq!(uris, expected);
@@ -409,6 +485,7 @@ $$
             text: "https://example.com/destination".to_string(),
             element: Some("a".to_string()),
             attribute: Some("href".to_string()),
+            span: span(1, 3),
         }];
         let uris = extract_markdown(markdown, true, true);
         assert_eq!(uris, expected);
@@ -422,11 +499,13 @@ $$
                 text: "https://example.com/destination".to_string(),
                 element: Some("a".to_string()),
                 attribute: Some("href".to_string()),
+                span: span(1, 3),
             },
             RawUri {
                 text: "https://example.com/source".to_string(),
                 element: Some("a".to_string()),
                 attribute: Some("href".to_string()),
+                span: span(1, 38),
             },
         ];
         let uris = extract_markdown(markdown, true, true);
@@ -439,4 +518,118 @@ $$
         let uris = extract_markdown(markdown, true, true);
         assert!(uris.is_empty());
     }
+
+    #[test]
+    fn test_link_text_not_checked() {
+        // Test that link text is not extracted as a separate link by default
+        let markdown =
+            r"[https://lycheerepublic.gov/notexist (archive.org link)](https://example.com)";
+        let uris = extract_markdown(markdown, false, false);
+
+        // Should only extract the destination URL, not the link text
+        let expected = vec![RawUri {
+            text: "https://example.com".to_string(),
+            element: Some("a".to_string()),
+            attribute: Some("href".to_string()),
+            span: span(1, 1),
+        }];
+
+        assert_eq!(uris, expected);
+        assert_eq!(
+            uris.len(),
+            1,
+            "Should only find destination URL, not link text"
+        );
+    }
+
+    #[test]
+    fn test_link_text_checked_with_include_verbatim() {
+        // Test that link text IS extracted when include_verbatim is true
+        let markdown =
+            r"[https://lycheerepublic.gov/notexist (archive.org link)](https://example.com)";
+        let uris = extract_markdown(markdown, true, false);
+
+        // Should extract both the link text AND the destination URL
+        let expected = vec![
+            RawUri {
+                text: "https://example.com".to_string(),
+                element: Some("a".to_string()),
+                attribute: Some("href".to_string()),
+                span: span(1, 1),
+            },
+            RawUri {
+                text: "https://lycheerepublic.gov/notexist".to_string(),
+                element: None,
+                attribute: None,
+                span: span(1, 2),
+            },
+        ];
+
+        assert_eq!(
+            uris.len(),
+            2,
+            "Should find both destination URL and link text"
+        );
+        // Check that both expected URLs are present (order might vary)
+        for expected_uri in expected {
+            assert!(
+                uris.contains(&expected_uri),
+                "Missing expected URI: {expected_uri:?}"
+            );
+        }
+    }
+
+    #[test]
+    fn test_reference_links_extraction() {
+        // Test that all types of reference links are extracted correctly
+        let markdown = r"
+Inline link: [link1](target1.md)
+
+Reference link: [link2][ref2]
+Collapsed link: [link3][]
+Shortcut link: [link4]
+
+[ref2]: target2.md
+[link3]: target3.md
+[link4]: target4.md
+";
+        let uris = extract_markdown(markdown, false, false);
+
+        let expected = vec![
+            RawUri {
+                text: "target1.md".to_string(),
+                element: Some("a".to_string()),
+                attribute: Some("href".to_string()),
+                span: span(2, 14),
+            },
+            RawUri {
+                text: "target2.md".to_string(),
+                element: Some("a".to_string()),
+                attribute: Some("href".to_string()),
+                span: span(4, 17),
+            },
+            RawUri {
+                text: "target3.md".to_string(),
+                element: Some("a".to_string()),
+                attribute: Some("href".to_string()),
+                span: span(5, 17),
+            },
+            RawUri {
+                text: "target4.md".to_string(),
+                element: Some("a".to_string()),
+                span: span(6, 16),
+                attribute: Some("href".to_string()),
+            },
+        ];
+
+        assert_eq!(uris.len(), 4, "Should extract all four link types");
+
+        // Check that all expected URIs are present (order might vary)
+        for expected_uri in expected {
+            assert!(
+                uris.contains(&expected_uri),
+                "Missing expected URI: {expected_uri:?}. Found: {uris:?}"
+            );
+        }
+    }
 }
